/*! \file   tm.c
    \brief  Task management
    \author Markus L. Noga <noga@inrialpes.fr>
    
    Contains the multitasking switcher and scheduler as
    well as library functions relating to task management.
*/
    
/*
 *  The contents of this file are subject to the Mozilla Public License
 *  Version 1.0 (the "License"); you may not use this file except in
 *  compliance with the License. You may obtain a copy of the License at
 *  http://www.mozilla.org/MPL/
 *
 *  Software distributed under the License is distributed on an "AS IS"
 *  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *  License for the specific language governing rights and limitations
 *  under the License.
 *
 *  The Original Code is legOS code, released March 19, 1999.
 *
 *  The Initial Developer of the Original Code is Markus L. Noga.
 *  Portions created by Markus L. Noga are Copyright (C) 1999
 *  Markus L. Noga. All Rights Reserved.
 *
 *  Contributor(s): Markus L. Noga <noga@inrialpes.fr>
 *                  Ben Laurie <ben@algroup.co.uk>
 *                  Lou Sortman <lou (at) sunsite (dot) unc (dot) edu>
 */

#include "sys/tm.h"

#ifndef NO_TASK_MANAGEMENT

#include "sys/mm.h"
#include "sys/time.h"
#include "sys/irq.h"
#include "sys/bitops.h"
#include "stdlib.h"
#include "unistd.h"

#ifndef NO_TASK_VISUALIZATION
# include "rom/lcd.h"
# include "conio.h"
#endif

///////////////////////////////////////////////////////////////////////////////
//
// Variables
//
///////////////////////////////////////////////////////////////////////////////

process_data pd_single;				//!< single process process data

process_data *cpid=&pd_single;			//!< ptr to current process data

priority_chain *priority_head;			//!< head of priority chain

process_data *pd_idle;				//!< idle proces

unsigned nb_tasks;				//!< number of tasks


///////////////////////////////////////////////////////////////////////////////
//
// Functions
//
///////////////////////////////////////////////////////////////////////////////

//! the task switcher
/*! saves active context and passes sp to scheduler
    then restores new context from returned sp
*/
void tm_switcher(void);
__asm__("
.text
.align 1
.globl _tm_switcher
_tm_switcher:
		; r6 saved by ROM
                ; r0 saved by system timer handler
    		
		mov.w	r1,@-r7			; save registers
		mov.w	r2,@-r7		
		mov.w	r3,@-r7		
		mov.w	r4,@-r7		
		mov.w	r5,@-r7		
		
		mov.w	r7,r0			; pass sp
		
		jsr	_tm_scheduler		; call scheduler
                
_tm_switcher_return:		
		mov.w	r0,r7			; set new sp
		
          	mov.w	@r7+,r5
		mov.w	@r7+,r4
		mov.w	@r7+,r3
		mov.w	@r7+,r2
		mov.w	@r7+,r1
		
                ; r0 will be restored by system timer handler
		; r6 will be restored by ROM

		rts				; return to new task
	");


//! the process scheduler
/*! \param old_sp current task's current stack pointer
    \return    new task's current stack pointer
  
    actual context switches performed by tm_switcher (assembler wrapper)
*/
size_t *tm_scheduler(size_t *old_sp) {
	process_data *next;			// next process to execute
	priority_chain	*priority;
	wakeup_t tmp;
	
#ifndef NO_TASK_VISUALIZATION
	static unsigned char sequence=0;
                
	if(cpid!=pd_idle) {
		sequence++;			// yeah we're running
		if(sequence==10) {
			lcd_show(man_stand);
			lcd_refresh();
		} else if(sequence==20) {
			lcd_show(man_run);
			lcd_refresh();
			sequence=0;
		}
	}
#endif	// NO_TASK_VISUALIZATION
                
	priority=cpid->priority;
	switch(cpid->pstate) {
	case P_ZOMBIE:				// did this one finish?
		// remove from queue
		if(cpid->next!=cpid)
		{
			// simply unlink the process
			priority->cpid =
			cpid->next->prev = cpid->prev;
			cpid->prev->next = cpid->next;
		} else {
			// priority chain is empty, remove it
			if(priority->prev != NULL)
				priority->prev->next = priority->next;
			else
				priority_head = priority->next;
			if(priority->next != NULL)
				priority->next->prev = priority->prev;
			free(priority);
		}
			
		// We're on that stack frame being freed right now,
		// but nobody can interrupt us anyways.
                //
		free(cpid->stack_base);	        // free stack
		free(cpid);		        // free process data
		
		//
		// FIXME: exit code?
		//
                
		switch(--nb_tasks) {
		case 1:	// only the idle process remains
			/* This code is the same as what is in kill()
			 * except that it does not mess with interrupts
			 * and we know that there is only one process.
			 */
			*( (priority_head->cpid->sp_save) + SP_RETURN_OFFSET )=
				(size_t) &exit;
			// in case it's waiting.
			priority_head->cpid->pstate=P_SLEEPING;
			break;
		case 0: // the last process has been removed
			// -> stop switcher,
    			systime_set_switcher(RTS_INSTRUCTION);
			cpid=&pd_single;	         // go single tasking.
			return cpid->sp_save;
		}
#ifndef NO_TASK_VISUALIZATION
		lcd_digit(nb_tasks);
		lcd_refresh();
#endif
		break;
		
	case P_RUNNING:
		cpid->pstate=P_SLEEPING;
		// no break
		
	case P_WAITING:
		cpid->sp_save=old_sp;
	}

	
	// find next process willing to run
	//	
	priority=priority_head;
	next=priority->cpid->next;
	for(;;)
	{
		if(next->pstate!=P_WAITING) break;
		if( ( tmp = next->wakeup(next->wakeup_data) ) != 0) {
			next->wakeup_data = tmp;
			break;
		}
		if(next == priority->cpid)	
		{	// if we've scanned the whole chain,
			// go to next priority
			if(priority->next != NULL) 
				priority = priority->next;
			else
				priority = priority_head;
			next=priority->cpid->next;
		} else {
			next=next->next;
		}
	}
	cpid=next->priority->cpid=next;	// execute next process
	cpid->pstate=P_RUNNING;

	return cpid->sp_save;
}


//! yield the rest of the current timeslice
/*! doesn't speed up the system clock.
*/
extern void yield(void);
__asm__("
.text
.globl _yield
.align 1
_yield:
		stc	ccr,r0h			; to fake an IRQ, we have to
		push	r0			; store the registers
                orc     #0x80,ccr               ; disable interrupts
    
                push    r6                      ; store r6
                
                mov.w   #0x04d4,r0              ; store rom return addr
                push    r0
                
                push    r0                      ; store r0 (destroyed by call.)

                mov.w   #_systime_tm_return,r0  ; store systime return addr
                push    r0

                jmp @_tm_switcher               ; call task switcher
	");
	

//! the idle task
/*! infinite sleep instruction to conserve power.
*/
extern int tm_idle_task(int argc,char **argv);
__asm__("
.text
.align 1
_tm_idle_task:
	   	sleep
		bra _tm_idle_task
	");
			
//! init task management
/*! called in single tasking mode before task setup.
*/
void tm_init(void) {
	// no tasks right now.
	//
	nb_tasks=0;
	
	// the single tasking context
	//
	pd_single.pstate=P_RUNNING;
	cpid=&pd_single;

	// the idle process is an institution
	//	
	pd_idle=(process_data*)execi(&tm_idle_task,0,NULL,0,IDLE_STACK_SIZE);

        systime_set_timeslice(TM_DEFAULT_SLICE);
}	


//! start task management 
/*! called in single tasking mode after task setup
*/
void tm_start(void) {
	disable_irqs();				// no interruptions, please

        systime_set_switcher(&tm_switcher);
	yield();				// go!
        
	enable_irqs();				// restored state would 
						// disallow interrupts
}

//! execute a memory image.
/*! \param code_start start address of code to execute
    \param argc       first argument passed, normally number of strings in argv
    \param argv       second argument passed, normally pointer to argument pointers.
    \param priority   new task's priority
    \param stack_size stack size for new process
    \return -1: fail, else pid.
    
    will return to caller in any case.
*/
pid_t execi(int (*code_start)(int,char**),int argc, char **argv,
            priority_t priority,size_t stack_size) {
	process_data *pd;
	priority_chain *pchain, *ppchain;	// for traversing priority chain
	priority_chain *newpchain;      	// for allocating new priority chain
	size_t *sp;

	// get memory
	//
	// task & stack area belong to parent process
	// they aren't freed by mm_reaper()
	//
	if((pd=malloc(sizeof(process_data)))==NULL)	
		return -1;
	
	if((sp=malloc(stack_size))==NULL) {
		free(pd);
		return -1;
	}

	pd->stack_base=sp;		// these we know already.
	
	sp+=(stack_size>>1);		// setup initial stack
				
	// when main() returns a value, it passes it in r0
	// as r0 is also the register to pass single int arguments by
	// gcc convention, we can just put the address of exit on the stack.
	
	*(--sp)=(size_t) &exit;
	
	// we have to construct a stack stub so tm_switcher,
        // systime_handler and the ROM routine can fill the 
        // right values on startup.
	
	*(--sp)=(size_t) code_start;	// entry point    < these two are for
	*(--sp)=0;			// ccr            < rte in ROM
	*(--sp)=0;			// r6             < pop r6 in ROM
	*(--sp)=(size_t)
                OCIA_ROM_RETURN;	// ROM return     < rts in systime_handler

       	*(--sp)=(size_t) argc;          // r0             < pop r0 in systime handler
        *(--sp)=(size_t)              
                &systime_tm_return;     // systime return < rts in tm_switcher

	*(--sp)=(size_t) argv;		// r1..r5	  < pop r1..r5 in tm_switcher
	*(--sp)=0;
	*(--sp)=0;
	*(--sp)=0;
	*(--sp)=0;
	
	pd->sp_save=sp;			// save sp for tm_switcher
	pd->pstate=P_SLEEPING;		// task is waiting for execution

	pd->parent=cpid;		// set parent

	// insert new process after all processes with higher priority
	// task queue manipulation -> lock irqs
			
	disable_irqs();

	ppchain=NULL;
	for(	pchain = priority_head;
		pchain != NULL && (pchain->priority) > priority;
		ppchain = pchain, pchain = pchain->next
		);
	if(pchain==NULL)
	{
		// add new chain at end
		pd->prev=pd->next=pd;
		if((newpchain=(priority_chain *)
			malloc(sizeof(priority_chain)))
			==NULL)
		{
			free(pd->stack_base);
			free(pd);
			enable_irqs();
			return -1;
		}
		newpchain->priority=priority;
		newpchain->prev=ppchain;
		newpchain->next=NULL;
		newpchain->cpid=pd;
		pd->priority=newpchain;
		if(ppchain != NULL)
			ppchain->next=newpchain;
		else
			priority_head=newpchain;
	} else {
		if(pchain->priority==priority)
		{
			// add process before the current one
			pd->prev = pchain->cpid->prev;
			pd->next = pchain->cpid;
			pd->prev->next =
			pchain->cpid->prev = pd;
			pd->priority = pchain;
		}
		else
		{
			// insert new chain before pchain
			pd->prev=pd->next=pd;
			if((newpchain=(priority_chain *)
				malloc(sizeof(priority_chain)))
				==NULL)
			{
				free(pd->stack_base);
				free(pd);
				enable_irqs();
				return -1;
			}
			newpchain->priority=priority;
			newpchain->prev=ppchain;
			newpchain->next=pchain;
			newpchain->cpid=pd;
			pd->priority=newpchain;
			if(ppchain != NULL)
				ppchain->next=newpchain;
			else
				priority_head=newpchain;
		}
	}

	nb_tasks++;
	enable_irqs();

#ifndef NO_TASK_VISUALIZATION
	lcd_digit(nb_tasks);		// display new number of tasks
#endif
							
	return (pid_t) pd;		// pid = (pid_t) &process_data_struct
}


//! exit task, returning code
/*! \param code The return code
 
    FIXME: for now, scrap the code.
*/
void exit(int code) {
	enable_irqs();				// just in case...
	mm_reaper();				// free dynamic memory
	cpid->pstate=P_ZOMBIE;			// come, task reaper!
	while(1)
		yield();
}


//! suspend process until wakeup function is non-null
/*! \param wakeup the wakeup function. called in task scheduler context.
    \param data argument passed to wakeup function by scheduler
    \return return value passed on from wakeup
*/
wakeup_t wait_event(wakeup_t (*wakeup)(wakeup_t),wakeup_t data) {
	cpid->wakeup     =wakeup;
	cpid->wakeup_data=data;
	cpid->pstate     =P_WAITING;
	
	yield();
	
	return cpid->wakeup_data;
}


//! wakeup function for sleep
/*! \param data time to wakeup encoded as a wakeup_t
*/
wakeup_t tm_sleep_wakeup(wakeup_t data) {
	return ((time_t)data)<=sys_time;
}

//! kill a process
/*! \param pid must be valid process ID, or undefined behaviour will result!
*/
void kill(pid_t pid) {
	process_data *pd=(process_data*) pid;
	if(pd==cpid)
		exit(-1);
	else {
		// when the task is switched to the next time,
		// make it exit immediatlely.

		disable_irqs();
		*( (pd->sp_save) + SP_RETURN_OFFSET )=(size_t) &exit;
		pd->pstate=P_SLEEPING;		// in case it's waiting.
		enable_irqs();
	}
	
}
#endif
